#include "ForthEvaluator.h"
#include "Token.h"

#include <algorithm>
#include <cctype>
#include <string>
#include <unordered_set>
#include <gsl/gsl>

using namespace std::literals::string_literals;

namespace forth {

template<typename Source>
class NarrowProxy final
{
    Source source;

public:
    explicit NarrowProxy(Source const &src): source{src}
    {}

    template<typename Target>
    operator Target() const
    {
        return gsl::narrow<Target>(source);
    }
};

template<typename Source>
NarrowProxy<Source> narrow(Source const &source)
{
    return NarrowProxy<Source>{source};
}

std::vector<std::string> splitAtWhiteSpace(std::string const &input) noexcept
{
    std::vector<std::string> result{};

    std::string temp{};
    for (auto const c : input)
    {
        if (!std::isspace(c))
        {
            temp += c;
        }
        else if (!temp.empty())
        {
            result.push_back(temp);
            temp.clear();
        }
    }
    if (!temp.empty())
    {
        result.push_back(temp);
    }

    return result;
}

static std::unordered_set<std::string> const keywords{":"s, ";"s};

std::vector<std::int16_t> evaluate(std::string const &source) noexcept(false)
{
    auto const words = splitAtWhiteSpace(source);
    std::vector<Token> tokens;
    tokens.reserve(words.size());
    std::transform(std::cbegin(words),
                   std::cend(words),
                   std::back_inserter(tokens),
                   [](std::string const &word) -> Token
                   {
                       try
                       {
                           return token::Number{gsl::narrow<std::int16_t>(std::stoi(word))};
                       }
                       catch (std::invalid_argument const &)
                       {
                           if (keywords.find(word) == keywords.end())
                           {
                               return token::Word{word};
                           }
                           else
                           {
                               return token::Keyword{word};
                           }
                       }
                   });

    std::vector<std::int16_t> stack;

    for (auto const &token : tokens)
    {
        if (token.IsNumber())
        {
            stack.push_back(token.GetNumber().value);
        }
        else if (token.IsWord())
        {
            auto const instruction = token.GetWord().value;
            if (instruction == "DROP")
            {
                stack.pop_back();
            }
            else if (instruction == "+")
            {
                auto const a = stack.back();
                stack.pop_back();
                auto const b = stack.back();
                stack.pop_back();
                stack.push_back(narrow(a + b));
            }
        }
    }

    return stack;
}

} // namespace forth
