#include "Token.h"

#include <ostream>
#include <utility>

namespace forth {

namespace token {

std::ostream &operator<<(std::ostream &os, Number const &number)
{
    return os << "Number[" << number.value << "]";
}

std::ostream &operator<<(std::ostream &os, Keyword const &keyword)
{
    return os << "Keyword[\"" << keyword.value << "\"]";
}

std::ostream &operator<<(std::ostream &os, Word const &word)
{
    return os << "Word[\"" << word.value << "\"]";
}

bool Number::operator==(Number const &rhs) const noexcept
{
    return value == rhs.value;
}

bool Number::operator!=(Number const &rhs) const noexcept
{
    return !operator==(rhs);
}

bool Keyword::operator==(Keyword const &rhs) const noexcept
{
    return value == rhs.value;
}

bool Keyword::operator!=(Keyword const &rhs) const noexcept
{
    return !operator==(rhs);
}

bool Word::operator==(Word const &rhs) const noexcept
{
    return value == rhs.value;
}

bool Word::operator!=(Word const &rhs) const noexcept
{
    return !operator==(rhs);
}

} // namespace token

Token::Token(token::Number value) noexcept: type{Type::Number}
{
    number = value;
}

Token::Token(token::Keyword value) noexcept: type{Type::Keyword}
{
    new (&keyword) token::Keyword{std::move(value)};
}

Token::Token(token::Word value) noexcept: type{Type::Word}
{
    new (&word) token::Word{std::move(value)};
}

Token::Token(Token const &src): type{src.type}
{
    switch (type)
    {
        case Type::Number:
            number = src.number;
            break;
        case Type::Keyword:
            new (&keyword) token::Keyword{src.keyword};
            break;
        case Type::Word:
            new (&word) token::Word{src.word};
    }
}

Token::Token(Token &&src) noexcept: type{src.type}
{

    switch (type)
    {
        case Type::Number:
            number = src.number;
            break;
        case Type::Keyword:
            new (&keyword) token::Keyword{std::move(src.keyword)};
            break;
        case Type::Word:
            new (&word) token::Word{std::move(src.word)};
    }
}

void Token::destroy() noexcept
{
    switch (type)
    {
        case Type::Number:
            destroy(&number);
            break;
        case Type::Keyword:
            destroy(&keyword);
            break;
        case Type::Word:
            destroy(&word);
    }
}

Token &Token::operator=(Token const &src)
{
    if (std::addressof(src) != this)
    {
        destroy();
        type = src.type;
        switch (type)
        {
            case Type::Number:
                number = src.number;
                break;
            case Type::Keyword:
                new (&keyword) token::Keyword{src.keyword};
                break;
            case Type::Word:
                new (&word) token::Word{src.word};
        }
    }
    return *this;
}

Token &Token::operator=(Token &&src) noexcept
{
    if (std::addressof(src) != this)
    {
        destroy();
        type = src.type;
        switch (type)
        {
            case Type::Number:
                number = src.number;
                break;
            case Type::Keyword:
                new (&keyword) token::Keyword{std::move(src.keyword)};
                break;
            case Type::Word:
                new (&word) token::Word{std::move(src.word)};
        }
    }
    return *this;
}

Token::~Token() noexcept
{
    destroy();
}

bool Token::IsNumber() const noexcept
{
    return type == Type::Number;
}

bool Token::IsKeyword() const noexcept
{
    return type == Type::Keyword;
}

bool Token::IsWord() const noexcept
{
    return type == Type::Word;
}

token::Number const &Token::GetNumber() const
{
    if (!IsNumber())
    {
        throw WrongTokenException{"Token is not a Number"};
    }
    return number;
}

token::Keyword const &Token::GetKeyword() const
{
    if (!IsKeyword())
    {
        throw WrongTokenException{"Token is not a Keyword"};
    }
    return keyword;
}

token::Word const &Token::GetWord() const
{
    if (!IsWord())
    {
        throw WrongTokenException{"Token is not a Word"};
    }
    return word;
}

bool Token::operator==(Token const &rhs) const noexcept
{
    return !operator!=(rhs);
}

bool Token::operator!=(Token const &rhs) const noexcept
{
    if (type != rhs.type)
    {
        return true;
    }

    switch (type)
    {
        case Type::Number:
            return number != rhs.number;
        case Type::Keyword:
            return keyword != rhs.keyword;
        case Type::Word:
            return word != rhs.word;
    }
}

std::ostream &operator<<(std::ostream &os, Token const &token)
{
    if (token.IsNumber())
    {
        return os << token.GetNumber();
    }
    else if (token.IsKeyword())
    {
        return os << token.GetKeyword();
    }
    else
    {
        return os << token.GetWord();
    }
}

} // namespace forth
